"""
Exhaust Fan ActionMappers
============================

This module contains classes to implement ActionMappers for controlling exhaust fan actuators in the environment.
"""

import gymnasium as gym
from typing import Any, Dict, List, Tuple, Optional
from eprllib.Agents.ActionMappers.BaseActionMapper import BaseActionMapper
from eprllib.Utils.observation_utils import get_actuator_name
from eprllib.Utils.annotations import override
from eprllib.Utils.agent_utils import get_agent_name

class CentralAgentActionMapper(BaseActionMapper):
    REQUIRED_KEYS = {
        "exhaust_fan_actuator": Tuple[str, str, str],
        "window_actuator": Tuple[str, str, str],
        "hvac_availability_actuator": Tuple[str, str, str]
    }
    
    def __init__(
        self,
        action_mapper_config: Dict[str, Any]
    ):
        """
        This class implements the Exhaust Fan actions.

        Args:
            action_mapper_config (Dict[str, Any]): The configuration of the action function.
            It should contain the following keys:
                - modes (List[float]): The flow factor to modify the maximum flow of the exhaust fan. The order 
                in the list corresponds with mode of the Fan, usually mode 0 is off and mode 1 has the lower 
                flow factor. No more than 11 modes are allowed.
                - exhaust_fan_actuator (Tuple[str, str, str]): The configuration for the exhaust fan actuator.
        """
        # Validate the config.
        # config_validation(action_mapper_config, self.REQUIRED_KEYS)
        
        super().__init__(action_mapper_config)
        
        self.agent_name: str = "None"
        self.modes: List[float] = [0, 0.0600, 0.1313, 0.2093, 0.2973]
        self.exhaust_fan_actuator: Optional[str] = None
        self.window_actuator: Optional[str] = None
        self.hvac_availability_actuator: Optional[str] = None
        
        # Check that all the elements in self.modes list are floats in the range [0, 1].
        for mode in self.modes:
            if mode < 0 or mode > 1:
                raise ValueError("The modes must be in the range [0, 1].")
    
    @override(BaseActionMapper)    
    def get_action_space_dim(self) -> gym.Space[Any]:
        """
        Get the action space of the environment.

        Returns:
            gym.Space: Action space of the environment.
        """
        
        return gym.spaces.MultiDiscrete([len(self.modes), 5, 2])
    
    @override(BaseActionMapper)
    def agent_to_actuator_action(self, action: Any, actuators: List[str]) -> Dict[str,Any]:
        """
        This method is used to transform the agent action to actuator dict action. Consider that
        one agent may manage more than one actuator.

        Args:
            action (Any): The agent action, normally an int of float.
            actuators (List[str]): The list of actuator names that the agent manage.

        Return:
            Dict[str, Any]: Dictionary with the actions for the actuators.
        """
        if self.agent_name is "None":
            self.agent_name = get_agent_name(actuators)
            self.exhaust_fan_actuator = get_actuator_name(
                self.agent_name,
                self.action_mapper_config['exhaust_fan_actuator'][0],
                self.action_mapper_config['exhaust_fan_actuator'][1],
                self.action_mapper_config['exhaust_fan_actuator'][2]
            )
            self.window_actuator = get_actuator_name(
                self.agent_name,
                self.action_mapper_config['window_actuator'][0],
                self.action_mapper_config['window_actuator'][1],
                self.action_mapper_config['window_actuator'][2]
            )
            self.hvac_availability_actuator = get_actuator_name(
                self.agent_name,
                self.action_mapper_config['hvac_availability_actuator'][0],
                self.action_mapper_config['hvac_availability_actuator'][1],
                self.action_mapper_config['hvac_availability_actuator'][2]
            )
            
        actuator_dict_actions: Dict[str, Any] = {actuator: None for actuator in actuators}
        
        assert isinstance(self.exhaust_fan_actuator, str), f"Exhaust fan actuator is not set. It should be a string, but got {type(self.exhaust_fan_actuator)}."
        assert isinstance(self.window_actuator, str), f"Window actuator is not set. It should be a string, but got {type(self.window_actuator)}."
        assert isinstance(self.hvac_availability_actuator, str), f"HVAC availability actuator is not set. It should be a string, but got {type(self.hvac_availability_actuator)}."
        
        actuator_dict_actions.update({
            self.exhaust_fan_actuator: self.modes[action[0]],
        })
        
        actuator_dict_actions.update({self.window_actuator: action[1] / 4})
        
        actuator_dict_actions.update({self.hvac_availability_actuator: action[2]})
        
        return actuator_dict_actions
    
    @override(BaseActionMapper)
    def get_actuator_action(self, action:float|int, actuator: str) -> Any:
        """
        This method is used to get the actions of the actuators after transform the 
        
        """
        return action



"""
Observation Utilities
======================

This module contain some utilities used within the observation space, function, and other parts
of the programm related with the observations.

"""

def get_variable_name(agent: str, variable_name:str, variable_key:str) -> str:
    """
    This function is used to get the variable name in the observation space.

    Args:
        agent (str): The agent name.
        variable_name (str): The variable name.
        variable_key (str): The variable key.

    Returns:
        str: The variable name in the observation space.
    """
    return f"{agent}: {variable_name}: {variable_key}"

def get_internal_variable_name(agent: str, variable_type:str, variable_key:str) -> str:
    """
    This function is used to get the internal variable name in the observation space.

    Args:
        agent (str): The agent name.
        variable_type (str): The variable type.
        variable_key (str): The variable key.

    Returns:
        str: The internal variable name in the observation space.
    """
    return get_variable_name(agent, variable_type, variable_key)

def get_meter_name(agent: str, meter_name:str) -> str:
    """
    This function is used to get the meter name in the observation space.

    Args:
        agent (str): The agent name.
        meter_name (str): The meter name.

    Returns:
        str: The meter name in the observation space.
    """
    return f"{agent}: {meter_name}"

def get_actuator_name(agent: str, actuator_component_type:str, actuator_control_type:str, actuator_key:str) -> str:
    """
    This function is used to get the actuator name in the observation space.

    Args:
        agent (str): The agent name.
        actuator_component_type (str): The actuator component type.
        actuator_control_type (str): The actuator control type.
        actuator_key (str): The actuator key.

    Returns:
        str: The actuator name in the observation space.
    """
    return f"{agent}: {actuator_component_type}: {actuator_control_type}: {actuator_key}"

def get_parameter_name(agent: str, parameter_name:str) -> str:
    """
    This function is used to get the parameter name in the observation space.

    Args:
        agent (str): The agent name.
        parameter_name (str): The parameter name.

    Returns:
        str: The parameter name in the observation space.
    """
    return f"{agent}: {parameter_name}"

def get_parameter_prediction_name(agent: str, parameter_name:str, hour:int) -> str:
    """
    This function is used to get the parameter prediction name in the observation space.

    Args:
        agent (str): The agent name.
        parameter_name (str): The parameter name.
        hour (int): The hour.

    Returns:
        str: The parameter prediction name in the observation space.
    """
    if hour < 1 or hour > 24:
        raise ValueError("ObservationUtils: Hour must be between 1 and 24.")
    return f"{agent}: {parameter_name}: +{hour}h"

def get_other_obs_name(agent: str, other_obs_name:str) -> str:
    """
    This function is used to get the other observation name in the observation space.

    Args:
        agent (str): The agent name.
        other_obs_name (str): The other observation name.

    Returns:
        str: The other observation name in the observation space.
    """
    return f"{agent}: {other_obs_name}"

def get_user_occupation_forecast_name(agent: str, hour:int) -> str:
    """
    This function is used to get the parameter prediction name in the observation space.

    Args:
        agent (str): The agent name.
        hour (int): The hour.

    Returns:
        str: The parameter prediction name in the observation space.
    """
    if hour < 1 or hour > 24:
        raise ValueError("ObservationUtils: Hour must be between 1 and 24.")
    return f"{agent}: User Occupation Forecast: +{hour}h"



"""
Windows Opening ActionMappers
================================

This module contains classes to implement window opening ActionMappers for controlling actuators in the environment.
"""
import gymnasium as gym
import numpy as np
from typing import Any, Dict, List, Tuple, Optional
from eprllib.Agents.ActionMappers.BaseActionMapper import BaseActionMapper
from eprllib.Utils.observation_utils import get_actuator_name
from eprllib.Utils.annotations import override

class WindowsOpeningDiscreteActionMapper(BaseActionMapper):

    @override(BaseActionMapper)
    def setup(self):
        # Here you have access to the self.action_mapper_config and self.agent_name
        
        # The name of the actuator that will be controlled by the agent. 
        # This will be obtained from the actuators_config in the environment configuration file.
        self.window_actuator: Optional[str] = None

        # Here we use the config dict to provide the action space dimension.
        self.action_space_dim: int = self.action_mapper_config.get("action_space_dim", 11)
    
    @override(BaseActionMapper)    
    def get_action_space_dim(self) -> gym.Space[Any]:
        """
        Get the action space of the environment.

        Returns:
            gym.Space: Action space of the environment.
        """
        return gym.spaces.Discrete(self.action_space_dim)
    
    @override(BaseActionMapper)
    def actuator_names(
        self, 
        actuators_config: Dict[str, Tuple[str,str,str]]
        ) -> None:
        """
        This method is used to transform the agent dict action to actuator dict action. Consider that
        one agent could manage more than one actuator. For that reason, it is important to transform the
        action dict to actuator dict actions.

        Args:
            action (Any): The action to be transformed.
            actuators (List[str]): List of actuators controlled by the agent.

        Returns:
            Dict[str, Any]: Transformed actions for the actuators.
        """
        
        # Here the name of the actuator is obtained from the actuators_config 
        # in the environment configuration file.
        self.window_actuator = get_actuator_name(
            self.agent_name,
            actuators_config["Window Opening Actuator"][0],
            actuators_config["Window Opening Actuator"][1],
            actuators_config["Window Opening Actuator"][2]
        )
        
        
    @override(BaseActionMapper)
    def _agent_to_actuator_action(self, action: Any, actuators: List[str]) -> Dict[str, Any]:
        """
        Transform the agent action to actuator action.

        Args:
            action (Any): The action to be transformed.
            actuators (List[str]): List of actuators controlled by the agent.

        Returns:
            Dict[str, Any]: Transformed actions for the actuators.
        """
        if self.window_actuator in actuators:
            return {self.window_actuator: action/10}
        else:
            return {}
    

class WindowsOpeningContinousActionMapper(BaseActionMapper):
    @override(BaseActionMapper)
    def setup(self):
        # Here you have access to the self.action_mapper_config and self.agent_name
        
        # The name of the actuator that will be controlled by the agent. 
        # This will be obtained from the actuators_config in the environment configuration file.
        self.window_actuator: Optional[str] = None
        
    
    @override(BaseActionMapper)    
    def get_action_space_dim(self) -> gym.Space[Any]:
        """
        Get the action space of the environment.

        Returns:
            gym.Space: Action space of the environment.
        """
        return gym.spaces.Box(low=0.0, high=1.0, shape=(1,), dtype=np.float32)
    
    
    @override(BaseActionMapper)
    def actuator_names(
        self, 
        actuators_config: Dict[str, Tuple[str,str,str]]
        ) -> None:
        """
        This method is used to transform the agent dict action to actuator dict action. Consider that
        one agent could manage more than one actuator. For that reason, it is important to transform the
        action dict to actuator dict actions.

        Args:
            action (Any): The action to be transformed.
            actuators (List[str]): List of actuators controlled by the agent.

        Returns:
            Dict[str, Any]: Transformed actions for the actuators.
        """
        
        # Here the name of the actuator is obtained from the actuators_config 
        # in the environment configuration file.
        self.window_actuator = get_actuator_name(
            self.agent_name,
            actuators_config["Window Opening Actuator"][0],
            actuators_config["Window Opening Actuator"][1],
            actuators_config["Window Opening Actuator"][2]
        )
        
        
    @override(BaseActionMapper)
    def _agent_to_actuator_action(self, action: Any, actuators: List[str]) -> Dict[str, Any]:
        """
        Transform the agent action to actuator action.

        Args:
            action (Any): The action to be transformed.
            actuators (List[str]): List of actuators controlled by the agent.

        Returns:
            Dict[str, Any]: Transformed actions for the actuators.
        """
        if self.window_actuator in actuators:
            return {self.window_actuator: action}
        else:
            return {}
    